# Management of Access Control Lists in Zope.
# The lists are derived from a DatabaseConnector.
# Copyright (C) 2000 Shane Hathaway
# E-mail: shathaway@earthling.net
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'''Functions of ACLManager:
- Link between GroupACL and user database
- Grants member list access to GroupACLs through the management interface
- Holds a requests for access to a group, enabling a
   sysadmin to authorize many GroupACL->group assignments at once.
- Manages some flags for each group, enabled by a sysadmin.

The DatabaseConnector must be in the same directory because
otherwise people could create ACLManager objects in folders
that access DatabaseConnector with full privileges.  Even if
they are prohibited from creating ACLManager objects, they
might create one using the Zope import mechanism.

DatabaseConnector "proxies" must not be allowed either.

The DatabaseConnector should support the following methods:
listGroupNames() : list of strings
listLocalEnrollment() :
  Returns list of MinimalUser (login, name, groupNames, roles, domains)
  objects.  Note that a user record is allowed to appear more than once in
  the returned list because the user is in multiple groups, for example.
  Also note that DatabaseConnector will call getLocalGroupNames(),
  available through acquisition in the virtual context, to
  discover the local group names.
getLocalUser(login) :
  Returns the given MinimalUser. May return multiple records.
checkPassword(login, password) :
  Returns 0 or 1.
'''

__version__ = '0.4'

from Globals import Persistent, HTMLFile, MessageDialog, PersistentMapping
import string, time
import OFS.SimpleItem
import Acquisition
import AccessControl

try:
    from Products.DatabaseAPI.DatabaseAPI \
         import _getPseudoVersion, _updatePseudoVersion
    from Products.DatabaseAPI.DatabaseConnector \
         import locateTrueParent, _listAvailablePeers
except:
    raise 'ImportError', \
          '''The Database API product is required for ACLManager and GroupACL.
          Please install it.'''


#def _debug(msg): print msg
def _debug(msg): pass

def manage_addACLManager(self, REQUEST):
    '''Adds an ACLManager to a folder.'''
    folder = self
    if REQUEST is not None:
        userdb_path = REQUEST['userdb_path']
        title = REQUEST['title']
        folder._setObject(ACLManager.id, ACLManager(title, userdb_path))
    return folder.manage_main(folder, REQUEST)

manage_addACLManagerForm = HTMLFile('aclManagerAdd', globals())

def listAvailableDatabases(self, includeParent):
    '''Looks for peer DatabaseConnector objects
    and returns a list of them as well as
    sub-DatabaseConnector objects.'''
    lst = _listAvailablePeers(self, includeParent, '_isADatabaseConnector')
    _augmentAvailableList(self, lst, 10, None)
    return lst

def _augmentAvailableList(self, lst, recurse_level, prepend):
    '''Augments the list with sub-databases.'''
    if recurse_level < 0:
        return lst
    idx = 0
    while idx < len(lst):
        id = lst[idx]
        obj = getattr(self, id)
        sublst = _listAvailablePeers(obj, 0, '_isADatabaseConnector')
        if len(sublst) > 0:
            if prepend:
                p = '%s/%s' % (prepend, id)
            else:
                p = id
            _augmentAvailableList(obj, sublst, recurse_level - 1, p)
            for idx2 in range(len(sublst)):
                sublst[idx2] = '%s/%s' % (p, sublst[idx2])
            lst[idx + 1:idx + 1] = sublst
            idx = idx + len(sublst) + 1
        else:
            idx = idx + 1

def locateACLManager(object, traversalRecord=None):
    '''Locates an ACLManager, starting with the given object.
    '''
    loop = 100
    while object is not None:
        if hasattr(object.aq_base, 'acl_manager'):
            aclm = object.acl_manager
            if hasattr(aclm, '_isAnACLManager') and \
               aclm._isAnACLManager:
                return aclm
        if traversalRecord is not None:
            traversalRecord.append(str(object.id))
        object = locateTrueParent(object)
        loop = loop - 1 # Don't get into an infinite loop.
        if loop < 0:
            raise 'ACLManagerError', 'Was looping indefinitely'
    if traversalRecord is not None:
        # The last entry is the "id" method of Application.
        # Remove it.
        try: del traversalRecord[-1]
        except: pass
    return None


class FlagDefinition:

    def __init__(self, name, letter, desc, isBoolean, isInteger,
                 defaultLocal, defaultGlobal):
        self.flagName = name
        self.flagLetter = letter
        self.flagDesc = desc
        self.flagIsBoolean = isBoolean
        self.flagIsInteger = isInteger
        self.flagDefaultLocal = defaultLocal
        self.flagDefaultGlobal = defaultGlobal

    def compareFlagValue(self, flags, compare):
        'Checks the status of our flag in an object'
        val = getattr(flags, self.flagName)
        if val == 0:
            return (compare == 0)
        elif val > 0:
            return (compare > 0)
        else:
            return (compare < 0)

    def getFlagValue(self, flags):
        return getattr(flags, self.flagName)

    def flagIsNotBlank(self, flags):
        flag = getattr(flags, self.flagName)
        if flag is None or flag == '' or flag == 0 or flag == -1:
            return 0
        return 1      # True.


class AbstractFlags (Persistent):
    '''Must be extended with a class that defines the
    flags class attribute.'''

    def __init__(self, isGlobal=0, defaultFlags=None):
        # Flag values: 0 = no, 1 = yes, -1 = default
        self.defaultFlags = defaultFlags
        for flag in self.flags:
            if isGlobal:
                initValue = flag.flagDefaultGlobal
            else:
                initValue = flag.flagDefaultLocal
            setattr(self, flag.flagName, initValue)

    def isAFlag(self, name):
        'Checks whether the given name refers to a flag.'
        for definition in self.flags:
            if name == definition.flagName:
                return 1
        return 0

    def _setFlagValue(self, name, value):
        flagDef = None
        for definition in self.flags:
            if name == definition.flagName:
                flagDef = definition
                break
        if flagDef is not None:
            if flagDef.flagIsInteger:
                value = string.atoi(str(value))
            if getattr(self, name) != value:
                setattr(self, name, value)
                return 1
        return 0

    def _getFlagValue(self, name):
        if not self.isAFlag(name):
            return None
        f = getattr(self, name)
        if f == -1:
            if self.defaultFlags is not None:
                return self.defaultFlags._getFlagValue(name)
            else:
                return 0
        else:
            return f


class ACLManagerFlags(AbstractFlags):

    flags = (FlagDefinition('allowLocalLoginPages', 'L',
                            'Allow local login pages', 1, 0, -1, 1),
             #FlagDefinition('allowCreateGuests', 'G',
             #               'Allow creation of guest accounts', 1, -1, 1),
             #FlagDefinition('maximumGuestTimeoutDays', 'T',
             #               'Maximum guest timeout in days', 0, -1, 0)
             FlagDefinition('cookieMode', 'C',
                            'Use cookies', 1, 0, -1, 1),
             FlagDefinition('cookieTimeout', 'O',
                            'Cookie timeout in seconds', 0, 1, -1, 14400),
             # 14400 seconds = four hours (default cookie timeout)
             )


class OneAssignment (Persistent):
    '''Represents the mapping of one group to one ACL.
    '''

    def __init__(self, approved, groupName):
        self.approved = approved
        self.groupName = groupName


class PathAssignments (Persistent):
    '''Represents the mapping of any number of groups to one ACL.
    '''

    def __init__(self, manager, path, flags):
        self.manager = manager
        self.path = path
        #if flags is None:
        #    flags = ACLManagerFlags(isGlobal=0, defaultFlags=???)
        self.flags = flags
        self.__asgns = PersistentMapping()
        # __asgns is a mapping of groupName -> OneAssignment
        self._approvalList = []

    def _approve(self, groupName, approval):
        '''Returns 1 if changed.
        '''
        changed = 0
        if approval:
            if self.__asgns.has_key(groupName):
                if not self.__asgns[groupName].approved:
                    self.__asgns[groupName].approved = 1
                    changed = 1
            else:
                self.__asgns[groupName] = OneAssignment(1, groupName)
                changed = 1
        else:
            if self.__asgns.has_key(groupName):
                del self.__asgns[groupName]
                changed = 1
        self._updateApprovalList()
        return changed

    def _changeName(self, groupName, newName):
        if self.__asgns.has_key(groupName):
            asgn = self.__asgns[groupName]
            asgn.groupName = newName
            self.__asgns[newName] = asgn
            del self.__asgns[groupName]
        self._updateApprovalList()

    def _request(self, groupName):
        '''Returns 1 if request added, 0 if already approved.
        '''
        if self.__asgns.has_key(groupName):
            asgn = self.__asgns[groupName]
            if asgn.approved:
                return 0
            else:
                return 1
        else:
            self.__asgns[groupName] = OneAssignment(0, groupName)
            return 1

    def _removeAssignments(self, groupNames):
        '''Returns number removed.'''
        count = 0
        for groupName in groupNames:
            try:
                del self.__asgns[groupName]
                count = count + 1
            except KeyError:
                pass
        self._updateApprovalList()
        return count

    def listAssignments(self):
        '''Returns the list of OneAssignment objects.
        '''
        return self.__asgns.values()

    def _updateApprovalList(self):
        lst = []
        for asgn in self.__asgns.values():
            if asgn.approved:
                lst.append(asgn.groupName)
        self._approvalList = lst

    def _getApprovalList(self):
        # The following two lines can be removed after version 0.1.
        if not hasattr(self, '_approvalList'):
            self._updateApprovalList()
        return self._approvalList

    def _prepForView(self, acl):
        if acl is not None:
            if acl.title:
                self._v_title = acl.title
            elif hasattr(acl, 'aq_parent') and \
                 hasattr(acl.aq_parent, 'title'):
                self._v_title = acl.aq_parent.title
            else:
                self._v_title = ''
            self._v_exists = 1
        else:
            self._v_title = ''
            self._v_exists = 0

    def getTitle(self):
        '''Returns the title of the GroupACL object, assuming
        _prepForView was called recently.'''
        return self._v_title

    def getExists(self):
        '''Returns a flag indicating whether the GroupACL object
        exists, assuming _prepForView was called recently.'''
        return self._v_exists


class ACLManager (Persistent,
                  OFS.SimpleItem.Item,
                  Acquisition.Implicit,
                  AccessControl.Role.RoleManager):
    '''A bridge between GroupACLs and user databases.
    '''

    _isAnACLManager = 1  # With an underscore so it can't be faked.
    id = 'acl_manager'
    meta_type = 'ACL Manager'

    manage_options=(
        {'label':'Assignments', 'action':'manage_asgns'},
        {'label':'Flags',       'action':'manage_flags'},
        {'label':'Properties',  'action':'manage_props'},
        {'label':'Security',    'action':'manage_access'},
        {'label':'Undo',        'action':'manage_UndoForm'},
        )

    __ac_permissions__=(
        ('View management screens',    ('manage_asgns',
                                        'manage_flags',
                                        'manage_props',
                                        'manage_main',
                                        'editPathForm',
                                        'getUserDbPath',
                                        'verifyUserDb',
                                        'getPathAssignment',
                                        'listPathAssignments',
                                        'listPathAssignmentsWithRequests',
                                        'listPathAssignmentsWithoutRequests',
                                        'listAvailableFlags',
                                        'getDefaultFlags',
                                        'listGroupNames',
                                        'manage_asgnsEdit',
                                        'manage_flagsEdit',
                                        'manage_propsEdit',
                                        'manage_deletePaths',
                                        'listAvailableDatabases2',
                                        )),
        ('Assign GroupACLs to groups', ('manage_addAssignment',
                                        )),
        )

    manage_asgns = HTMLFile('aclasgns', globals())
    manage_flags = HTMLFile('aclflags', globals())
    manage_props = HTMLFile('aclprops', globals())
    editPathForm = HTMLFile('editPathForm', globals())
    manage_main = manage_asgns

    def __init__(self, title, userdb_path):
        self.title = title
        self.__setUserDbPath(userdb_path)
        self.__paths = PersistentMapping()
        # __paths is a mapping of Folder path -> PathAssignments object
        self.__defaultFlags = ACLManagerFlags(isGlobal=1)

    def _setId(self, id):
        '''Controls the id of this object, which can not change.'''
        if id != ACLManager.id:
            raise 'ACLManagerError', MessageDialog(
                title='Invalid Id',
                message='Cannot change the id of an ACLManager',
                action='manage_main')
        # The DatabaseConnector might be in a different place now.
        # Force a re-fetch of the database object.
        _updatePseudoVersion()

    def __setUserDbPath(self, path):
        '''Changes userdb_id and subdb_list.
        '''
        userdb_id = ''
        subdb_list = []
        lst = string.split(path, '/')
        if len(lst) > 0:
            userdb_id = lst[0]
            if len(lst) > 1:
                subdb_list = lst[1:]
        self.__userdb_id = userdb_id
        self.__subdb_list = subdb_list

    def getUserDbPath(self):
        '''Returns the path to the database.
        '''
        self.__ensureNewVersion()
        userdb_path = self.__userdb_id
        for id in self.__subdb_list:
            if len(userdb_path) > 0:
                userdb_path = '%s/%s' % (userdb_path, id)
            else:
                userdb_path = id
        return userdb_path

    def __ensureNewVersion(self):
        '''Adds attributes to self that were not used in
        previous versions.
        '''
        try:
            dummyTest = self.__userdb_id
        except AttributeError:
            self.__setUserDbPath('')
    
    def _getDb(self):
        '''Retrieves the database object from the cache if
        appropriate, or gets it from
        DatabaseConnector._getIsolatedDatabase().  Does not
        apply acquisition to the database.'''
        try:
            if self._v_pseudoVersion == _getPseudoVersion():
                return self._v_isolDb
        except AttributeError:
            pass   # Not cached.
        trueParent = locateTrueParent(self)
        self.__ensureNewVersion()
        if hasattr(trueParent.aq_base, self.__userdb_id):
            dbSource = getattr(trueParent, self.__userdb_id)
            try:
                db = dbSource._getIsolatedDatabase()
                self._v_isolDb = db
                self._v_pseudoVersion = _getPseudoVersion()
                return db
            except:
                return None
        else:
            return None

    def _getSubDbList(self):
        '''To be called by GroupACL.
        '''
        return self.__subdb_list

    def __accessUserDb(self):
        '''Accesses the user database with a virtual context of self.'''
        db = self._getDb()
        if db is None:
            return None
        subdb = db._openInContextOf(self)
        subdb_list = self.__subdb_list
        if len(subdb_list) > 0:
            # userdb is the top-level database. Find
            # the database identified by subdb_list.
            for id in subdb_list:
                if hasattr(subdb.aq_base, id):
                    subdb = getattr(subdb, id)
                else:
                    return None
        return subdb

    def verifyUserDb(self):
        '''Verifies that the user database exists.'''
        db = self.__accessUserDb()
        if db is None:
            return 0
        else:
            return 1

    def __prepForView(self, ps):
        '''Ensures each PathAssignment in a list is updated with
        info about the GroupACL it represents.'''
        for p in ps:
            self.__prepOneForView(p)

    def __prepOneForView(self, p):
        '''Ensures a PathAssignment is updated with
        info about the GroupACL it represents.'''
        p._prepForView(self._getGroupACL(p.path))

    def getPathAssignment(self, path):
        '''Locates a PathAssignment.
        '''
        if self.__paths.has_key(path):
            rval = self.__paths[path]
            self.__prepOneForView(rval)
            return rval
        else:
            return None

    def listPathAssignments(self):
        '''Gets a list of all PathAssignments.'''
        rval = self.__paths.values()
        self.__prepForView(rval)
        return rval

    def listPathAssignmentsWithRequests(self, without=0):
        '''Gets a list of all PathAssignments with Assignments that
        are waiting for approval. If "without" is set, returns a list of
        PathAssignments without any Assignments waiting for
        approval.'''
        rval = []
        for pasgn in self.__paths.values():
            for asgn in pasgn.listAssignments():
                if not without and not asgn.approved or \
                   without and asgn.approved:
                    rval.append(pasgn)
                    break
        self.__prepForView(rval)
        return rval

    def listPathAssignmentsWithoutRequests(self):
        '''Gets a list of all PathAssignments with no Assignments
        waiting for approval.'''
        return self.listPathAssignmentsWithRequests(without=1)

    def listAvailableFlags(self):
        '''Lists the flags available in an ACLManagerFlags object.
        '''
        return ACLManagerFlags.flags

    def getDefaultFlags(self):
        '''Returns the default values of the flags.
        '''
        return self.__defaultFlags

    def listGroupNames(self):
        '''Lists the available group names from the database.
        '''
        db = self.__accessUserDb()
        if db is None:
            return None
        elif hasattr(db.aq_base, 'listGroupNames'):
            list = db.listGroupNames()
            # Polish the returned list.
            strType = type('')
            for idx in range(len(list)):
                item = list[idx]
                if type(item) != strType:
                    list[idx] = str(item)
            return list
        else:
            return None

    def __performAsgnEdits(self, REQUEST):
        '''Makes the changes listed in REQUEST.
        '''
        changes = 0
        # The strategy here is to search for all possible
        # field names, making changes only when the fields are
        # present.
        for pasgn in self.__paths.values():
            baseName = "%s." % pasgn.path
            fieldName = "%sdelete" % baseName
            if REQUEST.has_key(fieldName):
                # Remove this path.
                self.__logACLEvent(pasgn, \
                                   'all groups', 'Administrator removed')
                del self.__paths[pasgn.path]
                changes = changes + 1
            else:
                # Change the flags for pasgn.
                changes = changes + self.__performFlagEdits(
                    '%sflag' % baseName, pasgn.flags, REQUEST)
                # Approve, delete, and modify group assignments.
                for asgn in pasgn.listAssignments():
                    fieldName = '%sgroup.approval.%s' % \
                                (baseName, asgn.groupName)
                    if REQUEST.has_key(fieldName):
                        try:
                            approval = int(REQUEST[fieldName])
                            if approval == 0 or approval == 1:
                                if pasgn._approve(asgn.groupName, approval):
                                    if approval:
                                        event = 'Administrator approved'
                                    else:
                                        event = 'Administrator denied'
                                    self.__logACLEvent(pasgn, asgn.groupName, \
                                                       event)
                                    self._getGroupACL(pasgn.path \
                                                      )._updateDataVersion()
                                    changes = changes + 1
                        except ValueError:
                            pass  # Ignore.
                    fieldName = '%sgroup.name.%s' % \
                                (baseName, asgn.groupName)
                    if REQUEST.has_key(fieldName):
                        newName = REQUEST[fieldName]
                        if len(newName) > 0 and newName != asgn.groupName:
                            self.__logACLEvent(pasgn, newName, \
                                               'Administrator renamed')
                            pasgn._changeName(asgn.groupName, newName)
                            changes = changes + 1
                # Add groups manually.
                fieldName = "%saddgroup" % baseName
                if REQUEST.has_key(fieldName):
                    groupName = REQUEST[fieldName]
                    if len(groupName) > 0:
                        if pasgn._approve(groupName, 1):
                            self.__logACLEvent(pasgn, groupName, \
                                               'Administrator added')
                            self._getGroupACL(pasgn.path)._updateDataVersion()
                            changes = changes + 1
        return changes

    def __performFlagEdits(self, prefix, flags, REQUEST):
        '''Makes the changes listed in REQUEST.
        '''
        changes = 0
        for flagDef in ACLManagerFlags.flags:
            fieldName = '%s.%s' % (prefix, flagDef.flagName)
            if REQUEST.has_key(fieldName):
                val = REQUEST[fieldName]
                if flags._setFlagValue(flagDef.flagName, val):
                    changes = changes + 1
        return changes

    def __logACLEvent(self, pasgn, groupName, event):
        '''Writes to the GroupACL log, telling the user of
        approval or disapproval.
        '''
        groupACL = self._getGroupACL(pasgn.path)
        if groupACL is not None:
            ctime = time.ctime(time.time())
            logLine = '%s: %s \'%s\' -> \'%s\'\n' % \
                      (ctime, event, pasgn.path, groupName)
            groupACL.log = groupACL.log + logLine

    def manage_asgnsEdit(self, REQUEST):
        '''Accepts form fields and makes changes accordingly.
        '''
        changes = self.__performAsgnEdits(REQUEST)
        if changes == 1:
            msg = 'There was one change made.'
        else:
            msg = 'There were %d changes made.' % changes
        return MessageDialog(
                title='Changes accepted',
                message=msg,
                action='manage_asgns')

    def manage_deletePaths(self, selections, REQUEST):
        '''Deletes the selected PathAssignments.
        '''
        changes = 0
        for selection in selections:
            if self.__paths.has_key(selection):
                pasgn = self.__paths[selection]
                self.__logACLEvent(pasgn, \
                                   'all groups', 'Administrator removed')
                del self.__paths[selection]
                changes = changes + 1
        return MessageDialog(
                title='Deleted',
                message='%d path(s) removed.' % changes,
                action='manage_asgns')                
    
    def manage_flagsEdit(self, REQUEST):
        '''Accepts form fields and makes changes accordingly.
        '''
        changes = self.__performFlagEdits('defaultFlag',
                                          self.__defaultFlags, REQUEST)
        if changes == 1:
            msg = 'There was one change made.'
        else:
            msg = 'There were %d changes made.' % changes
        return MessageDialog(
                title='Changes accepted',
                message=msg,
                action='manage_asgns')

    def manage_propsEdit(self, REQUEST):
        '''Changes ACLManager fields.'''
        msg = 'Changes Accepted.'
        if REQUEST.has_key('title'):
            self.title = REQUEST['title']
        userdb_path = self.getUserDbPath()
        if REQUEST.has_key('userdb_path'):
            newpath = REQUEST['userdb_path']
            if newpath != userdb_path:
                self.__setUserDbPath(newpath)
                # Send a message to all GroupACL objects saying that
                # they should fetch the database object again.
                _updatePseudoVersion()
                
                if self.verifyUserDb():
                    msg = 'The database \'%s\' is ready.' % newpath
                else:
                    msg = 'The database \'%s\' is not ready ' \
                          'or was not found.' % newpath
        return MessageDialog(
            title='Changes accepted',
            message=msg,
            action='manage_asgns')

    def listAvailableDatabases2(self):
        '''Lists the DatabaseConnectors available to this object.
        '''
        return listAvailableDatabases(self, 1)

    def _getGroupACL(self, path):
        '''Locates and returns a GroupACL object by its path.
        '''
        ids = string.split(path, '/')
        obj = locateTrueParent(self)
        for id in ids:
            if hasattr(obj.aq_base, id):
                obj = getattr(obj, id)
            else:
                return None
        if hasattr(obj, '_isAGroupACL') and obj._isAGroupACL:
            return obj
        return None

    def _getApprovedGroupNames(self, path):
        '''Called by GroupACL, gets the list of group names
        approved for the given path.
        '''
        try:
            return self.__paths[path]._getApprovalList()
        except KeyError:
            pass
        return []

    def _getPathFlags(self, path):
        '''Called by GroupACL, gets the flags for the given path.
        '''
        paths = self.__paths
        try:
            return paths[path].flags
        except KeyError:
            pass
        return self.__defaultFlags

    def __getOrCreatePathAssignment(self, path):
        '''Gets or creates a specified PathAssignments object.
        '''
        if not self._getGroupACL(path):
            raise 'ACLManagerError', \
                  ('Path does not refer to a GroupACL: %s' % path)
        paths = self.__paths
        pasgn = None
        if paths.has_key(path):
            try:
                pasgn = paths[path]
            except KeyError:
                pass
        if pasgn is None:
            flags = ACLManagerFlags(defaultFlags=self.__defaultFlags)
            pasgn = PathAssignments(self, path, flags)
            paths[path] = pasgn
        return pasgn

    def manage_addAssignment(self, path, groupName, returnToACL=1):
        '''Adds an assignment, bypassing the request mechanism,
        then optionally sends the user back to the GroupACL object.
        Only available to users with the proper permission.
        '''
        pasgn = self.__getOrCreatePathAssignment(path)
        # We could check the validity of the group name here.
        pasgn._approve(groupName, 1)
        self.__logACLEvent(pasgn, groupName, 'Administrator assigned')
        self._getGroupACL(path)._updateDataVersion()
        if returnToACL:
            act = '../%s/manage_main' % path
        else:
            act = 'manage_asgns'            
        return MessageDialog(
            title='Group assignment added',
            message='Group assignment added: \'%s\' to \'%s\'' % \
            (path, groupName), \
            action=act)

    def _requestAssignment(self, path, groupName):
        '''Requests a path assignment. Returns 1 if created, 0
        if already approved.
        '''
        pasgn = self.__getOrCreatePathAssignment(path)
        # We could check the validity of the group name here.
        success = pasgn._request(groupName)
        if success:
            self.__logACLEvent(pasgn, groupName, 'Requested')
        return success

    def _removeAssignments(self, path, groupNames):
        '''Removes the given assignments from the GroupACL.
        Returns number removed.
        '''
        try:
            pasgn = self.__paths[path]
        except KeyError:
            return 0
        return pasgn._removeAssignments(groupNames)
