# A version of BasicCookieJar that stores cookies in
# Zope's "var" directory.  Maintenance is automatic.
# Copyright (C) 2000 Shane Hathaway
# E-mail: shathaway@earthling.net
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import os, string, sys, time
# import App.FindHomes
from BasicCookieJar import BasicCookieJar #, Cookie
from zLOG import LOG, ERROR, WARNING

try:
    from cPickle import dump, load
except:
    from Pickle import dump, load

varDir = os.path.join(INSTANCE_HOME, 'var', 'cookies')


def safeCookieName(cookieName,
                   safeChars="%s%s_" % (string.letters, string.digits)):
    cn = cookieName
    for i in range(len(cn)):
        if string.find(safeChars, cn[i]) < 0:
            cn[i] = '_'
    return cn


class BasicFSCookieJar (BasicCookieJar):

    # Note: remember that storageSubdir (below) is not intended to
    # be set by users.  IOW If a BasicFSCookieJar object is added to
    # a Zope folder, the user must not have the ability to set
    # storageSubdir while doing it.  Otherwise the security of
    # the ACLManager product would be compromised, along with
    # creating a number of other filesystem-related hazards.
    def __init__(self, cleanupInterval=1800, storageSubdir='pub'):
        BasicCookieJar.__init__(self, cleanupInterval)
        self.storageSubdir = storageSubdir
        self.__ensureStorageExists()

    def __getStorageDir(self):
        try:
            return self._v_storageDir
        except AttributeError:
            storageDir = os.path.join(varDir, self.storageSubdir)
            self._v_storageDir = storageDir
            return storageDir

    def __ensureStorageExists(self):
        storageDir = self.__getStorageDir()
        if not os.path.isdir(storageDir):
            os.makedirs(storageDir)

    def _cleanupPersistentStorage(self):
        storageDir = self.__getStorageDir()
        self.__ensureStorageExists()
        now = time.time()
        list = os.listdir(storageDir)
        for filename in list:
            p = os.path.join(storageDir, filename)
            if os.path.isdir(p):
                try:
                    if self.__cleanupCookieDir(p, now):
                        # Empty cookie directory.
                        os.rmdir(p)
                except:
                    LOG('BasicFSCookieJar', ERROR,
                        'Unable to clean directory %s: %s' \
                        % (p, str(sys.exc_info())), '', sys.exc_info())

    def __cleanupCookieDir(self, path, now):
        'Returns 1 if all cookies are removed.'
        cleanedAll = 1
        list = os.listdir(path)
        for filename in list:
            p = os.path.join(path, filename)
            if os.path.isfile(p):
                try:
                    if not self.__cleanupCookie(p, now):
                        cleanedAll = 0
                except:
                    LOG('BasicFSCookieJar', ERROR,
                        'Unable to clean cookie %s: %s' \
                        % (p, str(sys.exc_info())), '', sys.exc_info())
                    cleanedAll = 0
        return cleanedAll

    def __cleanupCookie(self, path, now):
        'Returns 1 if cookie is removed.'
        try:
            file = open(path, 'rb')
            c = load(file)
            file.close()
        except:
            # Cookie is being written by another process?
            return 1
        expiration = self._getInternalExpiration(c.cookieName, c.hash)
        if expiration > now:
            # Didn't expire in memory. May update expiration on disk.
            if expiration > c.expiration:
                # Put the latest expiration on disk.
                c.expiration = expiration
                file = open(path, 'w+b')
                dump(c, file, 1)
                file.close()
            else:
                # Get the latest expiration from disk.
                self._storeCookieInHashtable(c)
        else:
            # Expired in memory.
            if c.expiration <= now:
                # Expired on disk.
                os.remove(path)
                return 1
            else:
                # Load from disk.
                self._storeCookieInHashtable(c)
        return 0

    def __cookiePath(self, cookieName, hash, createDir):
        storageDir = self.__getStorageDir()
        cn = safeCookieName(cookieName)
        path = os.path.join(storageDir, cn)
        if createDir and not os.path.isdir(path):
            os.mkdir(path)
        return os.path.join(path, hash)

    def _getPersistentCookie(self, cookieName, hash):
        '''Retrieves a Cookie object.'''
        path = self.__cookiePath(cookieName, hash, 0)
        if path is not None and os.path.isfile(path):
            try:
                file = open(path, 'rb')
                c = load(file)
                file.close()
                return c
            except:
                LOG('BasicFSCookieJar', ERROR, \
                    'Unable to read cookie %s/%s: %s' \
                    % (cookieName, hash, str(sys.exc_info())), '', \
                    sys.exc_info())
                # (Don't try to delete it--we may be able to read it later.)
        return None

    def _setPersistentCookie(self, c):
        '''Stores a cookie value.'''
        path = self.__cookiePath(c.cookieName, c.hash, 1)
        try:
            file = open(path, 'w+b')
            dump(c, file, 1)
            file.close()
        except:
            LOG('BasicFSCookieJar', ERROR, \
                'Unable to write cookie %s/%s: %s' \
                % (cookieName, hash, str(sys.exc_info())), '', sys.exc_info())
            # (Don't try to delete it--another process might be writing it.)

    def _deletePersistentCookie(self, cookieName, hash):
        path = self.__cookiePath(cookieName, hash, 0)
        try:
            os.remove(path)
        except:
            # Ignore.
            pass
