# Management of Access Control Lists in Zope.
# The lists are derived from a DatabaseConnector.
# Copyright (C) 2000 Shane Hathaway
# E-mail: shathaway@earthling.net
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'''GroupACL.py
'''

__version__ = '0.4'

from Globals import Persistent, HTMLFile, MessageDialog, PersistentMapping
import string, time, base64
import Acquisition
import AccessControl
from AccessControl.User import BasicUserFolder, _remote_user_mode
from ACLManager import locateACLManager, locateTrueParent, \
     _getPseudoVersion, _updatePseudoVersion
from BasicFSCookieJar import BasicFSCookieJar
from User import User
from AccessControl.PermissionRole import _what_not_even_god_should_do
from OFS.Folder import Folder
from zLOG import LOG, INFO, ERROR, WARNING


#def _debug(msg): print msg
def _debug(msg): pass

def _augmentList(dest, src):
    '''Augments the contents of the dest list with
    the contents of the src list.
    '''
    toAppend = []
    if src:
        for item in src:
            if item not in dest:
                toAppend.append(item)
    if len(toAppend) > 0:
        return dest + toAppend
    else:
        return dest

def _concatMinimalUser(dest, src):
    '''Adds the contents of a MinimalUser object to a MinimalUser object.
    '''
    dest.groupNames = _augmentList(dest.groupNames, src.groupNames)
    dest.roles = _augmentList(dest.roles, src.roles)
    dest.domains = _augmentList(dest.domains, src.domains)
    
# The _privateCookieJar object will be created upon starting Zope.
# It will be common to all GroupACL instances.

_privateCookieJar = BasicFSCookieJar(storageSubdir='acls')
cname = '__acl_user'
nameparm = cname + '_name'
passparm = cname + '_password'

def manage_addGroupACL(self, title, REQUEST):
    '''Adds a GroupACL to a folder. Doesnt actually check whether an
    acl_manager object exists.'''
    folder = self
    if REQUEST is not None:
        folder._setObject(GroupACL.id, GroupACL(title))
    return folder.manage_main(folder, REQUEST)

manage_addGroupACLForm = HTMLFile('groupACLAdd', globals())


class UserCookie:
    '''This is what gets stored in the cookie jar.  There will be
    more information in memory than is in persistent storage.
    '''
    def __init__(self, authenticatedLogin):
        self.authenticatedLogin = authenticatedLogin
        # Note that although this object makes changes to itself,
        # the changes are never written to persistent storage
        # because of the way CookieJars work.  This is desirable
        # behavior, since the cache can be re-created easily.

    def __prepCachedUsers(self):
        if not hasattr(self, 'cachedUsers'):
            self.cachedUsers = {}

    def getUserInGroup(self, groupId):
        self.__prepCachedUsers()
        try:
            u = self.cachedUsers[groupId]
            return u
        except KeyError:
            return None

    def setUserInGroup(self, groupId, user):
        self.__prepCachedUsers()
        self.cachedUsers[groupId] = user


class DummyClass: pass

class GroupACL(Persistent,
               BasicUserFolder,
               Acquisition.Implicit,
               AccessControl.Role.RoleManager):
    '''A user folder derived from an ACLManager, which in turn
    is derived from a database.
    '''

    meta_type = 'Group ACL'
    id = 'acl_users'
    _isAGroupACL = 1

    manage_options = (
        {'label':'Group assignments', 'action':'manage_groups'},
        {'label':'Users',             'action':'manage_users'},
        {'label':'Properties',        'action':'manage_props'},
        {'label':'Security',          'action':'manage_access'},
        {'label':'Undo',              'action':'manage_UndoForm'},
        )

    __ac_permissions__ = (
        ('View management screens', ('manage_main',
                                     'manage_groups',
                                     'manage_flushCache',
                                     'getACLManager',
                                     'getTraversalPath',
                                     'getLocalGroupNames',
                                     'getFlagValue',
                                     'getUsersInfoList',
                                     'getGroupInfoList',
                                     'testAssignmentPermission',
                                     'getUser',
                                     )),
        ('Manage users',            ('manage_requestAssignment',
                                     'manage_removeAssignments',
                                     'manage_clearLog',
                                     #'manage_editUser',
                                     #'manage_editUserForm',
                                     'manage_users',
                                     'manage_addGroupForm',
                                     )),
        ('Manage properties',       ('manage_props',
                                     'manage_propsEdit')),
        ('Can Login and Logout',    ('login',
                                     'logout',
                                     'loginSuccess'),
                                      ('Anonymous', 'Manager',)),
        ('Use Database Methods',    ('db',),
                                      ('Anonymous', 'Manager',)),
        )

    manage_groups = HTMLFile('grpgroups', globals())
    manage_users = HTMLFile('grpusers', globals())
    manage_props = HTMLFile('grpprops', globals())
    manage_addGroupForm = HTMLFile('grpadd', globals())
    manage_main = manage_groups
    #manage_editUserForm = HTMLFile('editUser', globals())

    def __init__(self, title):
        self.__userInfo = PersistentMapping()
        self.title = title
        self.log = ''

    def _setId(self, id):
        BasicUserFolder._setId(self, id)
        # We may have moved, so we have to update our
        # connection to the ACL Manager.
        self._v_pseudoVersion = 0
    
    def getACLManager(self):
        '''Locates and returns the acl_manager object.  Note that
        the ACL Manager object has to be thread specific because
        it will acquire differently for each REQUEST.
        '''
        return locateACLManager(self.aq_parent)

    def __ensureCurrentConfig(self):
        '''Ensures db() has been called before
        accessing the non-thread-specific variables.
        '''
        try:
            if self._v_pseudoVersion != _getPseudoVersion():
                self._configure()
        except AttributeError:
            self.__configure()

    def __configure(self):
        '''Sets up some non-thread-specific variables from
        the ACL Manager and stores them without making them
        persistent.
        '''
        traversalRecord = []
        mng = locateACLManager(self.aq_parent, traversalRecord)
        traversalPath = str(self.id)
        for item in traversalRecord:
            traversalPath = "%s/%s" % (str(item), traversalPath)
        self._v_traversalPath = traversalPath
        if mng is not None:
            self._v_flags = mng._getPathFlags(traversalPath)
            self._v_approvedGroups = mng._getApprovedGroupNames(traversalPath)
            self._v_db = mng._getDb()
            self._v_subdb_list = mng._getSubDbList()
        else:
            self._v_flags = None
            self._v_approvedGroups = []
            self._v_db = None
            self._v_subdb_list = []
        # Reset pseudoVersion last--important for
        # multithreaded environments.
        self._v_pseudoVersion = _getPseudoVersion()

    def db(self):
        '''Gets the database object already set up
        to acquire from acl_users (this object).
        '''
        self.__ensureCurrentConfig()
        db = self._v_db
        if db is None:
            return None
        else:
            return db._openInContextOf(self)

    def userdb(self):
        '''Gets the user database object already set up
        to acquire from acl_users (this object).
        '''
        subdb = self.db()
        if subdb is None:
            return None
        # userdb is the top-level database. Find
        # the database identified by subdb_list.
        for id in self._v_subdb_list:
            if hasattr(subdb.aq_base, id):
                subdb = getattr(subdb, id)
            else:
                return None
        return subdb

    def getTraversalPath(self):
        '''Gets the true path from acl_manager to self.
        '''
        self.__ensureCurrentConfig()
        return self._v_traversalPath

    def getLocalGroupNames(self):
        '''Fetches the list of approved group names from ACLManager.
        A lot of the security of GroupACL depends on the correct
        passing of the list of group names to the database.
        '''
        self.__ensureCurrentConfig()
        return self._v_approvedGroups
    
    def getFlagValue(self, name):
        '''Returns the value of a flag set by a sysadmin.'''
        self.__ensureCurrentConfig()
        flags = self._v_flags
        if flags is None:
            return None
        else:
            rval = flags._getFlagValue(name)
            return rval

    def __mergeUserInfo(self):
        '''Updates self.__userInfo with data from the database
        and returns a dictionary containing cleaned-up User
        objects.
        '''
        userdb = self.userdb()
        if userdb == None or not hasattr(userdb.aq_base, 'listLocalEnrollment'):
            return {}
        dbList = userdb.listLocalEnrollment()

        # Merge duplicate records from dbList.
        dbDict = {}
        for dbUi in dbList:
            if dbDict.has_key(dbUi.login):
                dbUiOrig = dbDict[dbUi.login]
                _concatMinimalUser(dbUiOrig, dbUi)
            else:
                dbDict[dbUi.login] = dbUi
        # Link the user info with the localized custom user info.
        luiList = self.__userInfo
        rval = {}
        for dbUi in dbDict.values():
            if luiList.has_key(dbUi.login):
                # Merge data from our local list.
                lui = luiList[dbUi.login]
                lui._importUserInfo(dbUi)
                rval[dbUi.login] = lui
            else:
                # Not listed.  Add to the returned dictionary.
                lui = User(dbUi)
                rval[dbUi.login] = lui
        return rval

    def getUserNames(self):
        '''Returns a list of user names.'''
        dict = self.__mergeUserInfo()
        return dict.keys()

    def getUserInfoList(self):
        '''Returns a list of User objects.'''
        dict = self.__mergeUserInfo()
        return dict.values()

    def getGroupInfoList(self):
        '''Returns helpful information about each group name
        approved for this GroupACL, including name and userCount.'''
        userDict = self.__mergeUserInfo()
        groupDict = {}
        groupNames = self.getLocalGroupNames()
        # Create an object for each group and put it
        # in groupDict.
        for groupName in groupNames:
            groupInfo = DummyClass()
            groupInfo.name = groupName
            groupInfo.userCount = 0
            groupDict[groupName] = groupInfo
        # Count the number of users in each group.
        for user in userDict.values():
            ugn = user.getGroupNames()
            for gn in ugn:
                try:
                    groupInfo = groupDict[gn]
                    groupInfo.userCount = groupInfo.userCount + 1
                except KeyError:
                    # Rare, unimportant occurrence--probably a thread conflict.
                    pass
        # Return the list of group info objects.
        return groupDict.values()

    def testAssignmentPermission(self, REQUEST):
        '''Tests whether the user will be able to call manage_addAssignment
        successfully.  If not, the user can only make a request.
        This is not a real security check, otherwise it would call
        verify_watermark().
        '''
        user = REQUEST['AUTHENTICATED_USER']
        mng = self.getACLManager()
        if mng is None:
            return 0
        method = mng.manage_addAssignment
        return user.allowed(method, method.__roles__)

    def manage_requestAssignment(self, groupName):
        '''Enters a request into the ACLManager list.
        '''
        aclmng = self.getACLManager()
        if aclmng is None:
            raise 'GroupACLException', 'ACLManager not available.'
        if aclmng._requestAssignment(self.getTraversalPath(), groupName):
            msg = 'Request submitted.'
        else:
            msg = 'The requested assignment has already been approved.'
        return MessageDialog(            
            title='Group assignment added',
            message=msg,
            action='manage_groups')

    def manage_removeAssignments(self, selections):
        '''Allows the user to remove this GroupACL from groups.
        This is never prohibited by ACLManager.
        '''
        aclmng = self.getACLManager()
        if aclmng is None:
            raise 'GroupACLException', 'ACLManager not available.'
        count = aclmng._removeAssignments(self.getTraversalPath(), \
                                          selections)
        self._updateDataVersion()
        return MessageDialog(
            title='Group assignments',
            message='%d group assignment(s) removed.' % count,
            action='manage_groups')

    def manage_clearLog(self, REQUEST):
        '''Clears the log.'''
        self.log = ''
        return self.manage_groups(self, REQUEST)

    #def manage_editUser(self, login, extraRoles=[]):
    #    '''Modifies a user.'''
    #    user = self.getUser(login)
    #    if user is None:
    #        raise 'GroupACLException', 'User not found.'
    #    # Verify the user really can grant these roles.
    #    # Users are only allowed to grant roles they
    #    # themselves have.
    #    try:
    #        REQUEST = getattr(self, 'REQUEST')
    #        auth_user = REQUEST['AUTHENTICATED_USER']
    #        verify_watermark(auth_user)
    #    except:
    #        raise 'GroupACLException', 'Unable to locate AUTHENTICATED_USER.'
    #    for role in extraRoles:
    #        if not auth_user.has_role(role):
    #            raise 'GroupACLException', 'Cannot grant specified roles.'
    #    user.extraRoles = extraRoles
    #    # Ensure changes are committed to customization list.
    #    self.__userInfo[user.login] = user
    #    self._updateDataVersion()
    #    return MessageDialog(
    #        title='Edit user',
    #        message='Changes accepted.',
    #        action='manage_users')

    def _updateDataVersion(self):
        '''Causes all User objects for this folder stored in cookies
        to be invalidated.  Useful when updating roles or groups.
        '''
        self._v_dataVersion = time.time()
        self._v_basicUserCache = {}
        self._v_pseudoVersion = 0  # Also reconfigure.

    def _getDataVersion(self):
        '''Gets the pseudo version number for the User objects.
        '''
        try:
            dv = self._v_dataVersion
        except AttributeError:
            # Create the _v_dataVersion attribute.
            self._v_dataVersion = time.time()
            dv = self._v_dataVersion
        return dv

    def manage_flushCache(self, REQUEST):
        '''See _updateDataVersion().  This is very safe.
        '''
        self._updateDataVersion()
        return self.manage_groups(self, REQUEST)

    def getUser(self, name):
        '''Returns the named User object or None.'''
        
        # Special case for the superuser.
        if name == self._super.getUserName():
            return self._super
        
        userdb = self.userdb()
        if userdb != None and hasattr(userdb.aq_base, 'getLocalUser'):
            luiList = self.__userInfo
            dbUi = userdb.getLocalUser(name)
            if not dbUi:
                # Not found in database.
                return None
            else:
                # Combine all returned objects into one.
                combo = dbUi[0]
                for item in dbUi[1:]:
                    _concatMinimalUser(combo, item)
                dbUi = combo
                if luiList.has_key(name):
                    # Merge with the old info.
                    lui = luiList[name]
                    lui._importUserInfo(dbUi)
                    return lui.__of__(self)
                else:
                    # Create a user info object from the db and return it.
                    lui = User(dbUi)
                    return lui.__of__(self)
        else:
            # Database not available. Allow no one.
            return None

    def __readCookie(self, REQUEST):
        '''Returns the User object that applies to this
        GroupACL, either from the cookie cache or by creating it.
        '''
        c = _privateCookieJar.getCookie(REQUEST, cname)
        if c is None:
            return None
        else:
            user = c.getUserInGroup(self._p_oid)
            if user is not None:
                try:
                    if user._v_dataVersion != self._v_dataVersion:
                        # The User object was created before the last
                        # cache flush.  Create it again.
                        user = None
                    else:
                        user = user.__of__(self)
                except AttributeError:
                    # User was read from persistent storage.
                    user = None
            if user is None:
                user = self.getUser(c.authenticatedLogin)
                if user is not None:
                    # Cache the newly created User object.
                    user._v_dataVersion = self._getDataVersion()
                    c.setUserInGroup(self._p_oid, user.aq_base)
            return user

    def __setCookie(self, REQUEST, user):
        '''Sets a system-wide cookie that allows the user to
        access anywhere he/she is listed in the groups.
        '''
        if hasattr(user, 'aq_base'):
            user = user.aq_base
        c = UserCookie(user.login)
        _privateCookieJar.setCookie(REQUEST, cname, c,
                                    self.getFlagValue('cookieTimeout'))
        user._v_dataVersion = self._getDataVersion()
        c.setUserInGroup(self._p_oid, user)
        
    def manage_propsEdit(self, title, REQUEST):
        '''Changes the properties of this object.'''
        self.title = title
        return MessageDialog(
            title='Edit properties',
            message='Changes accepted.',
            action='manage_main')

    # The following methods (_cookie_validate,
    # _basic_validate, and validate) had their origins in
    # GenericUserFolder.py.  A great big thank you goes out to
    # Stuart "Zen" Bishop!  (zen@cs.rmit.edu.au)

    def _cookie_validate(self, request, auth, roles, parent):
	'''
	First attempt basic auth (for non-cookie protocols),
	and then try authenticating via form inputs from our login
	screen or an authentication token passed via cookie
	'''

	# If we have been authenticated using basic authentication
	# using a different acl_users folder, we can try authenticating
	# here with the same username & password.
	# This handles non-cookie connections like FTP.
	if auth and string.lower(auth[:6]) == 'basic ':
	    user = self._basic_validate(request, auth, roles, parent)
	    if user is not None:
		return user

        response=request['RESPONSE']

        username = None
        password = None
        hash = None
        blanked = 'This space intentionally blank'

        if request.form.has_key(nameparm):
            #
            # Pull username & cleartext from form input.
            # Ensure cleartext isn't left in the REQUEST for
            # malicious DTML
            #
            username = request[nameparm]
            request.form[nameparm] = blanked
            request[nameparm] = blanked

        if request.form.has_key(passparm):
            password = request[passparm]
            request.form[passparm] = blanked
            request[passparm] = blanked
        
        e = request.environ
        if _remote_user_mode and e.has_key('REMOTE_USER'):
            # Overrides remote user name.
            username = e['REMOTE_USER']
    
        if username is not None or password is not None:
            # Try to log in.
            # If user is logging in again, nuke their
            # old cookie right now
            _privateCookieJar.deleteCookie(request, cname)

	    # Try validating with username & password.
            user = self.getUser(username)
            if user is not None and \
               user.authenticate(password, request) and \
               user.allowed(parent, roles):
                self.__setCookie(request, user)
                LOG('GroupACL', INFO, \
                    '%s: User %s authenticated successfully (cookie).' % \
                    (self.getTraversalPath(), user.getUserName()))
                return user

        elif request.has_key(cname):
            # Try to validate using the encoded cookie presented by the user.
            user = self.__readCookie(request)
            if user is not None:
                if user.allowed(parent, roles):
                    return user

        # Replace the unauthorized method in response, enabling
        # the display of the login page.  An excellent solution
        # to an old problem.  Thanks Zen!
        if not hasattr(response, 'unauthorized') or \
           not hasattr(response.unauthorized, 'cookie_aware'):
            response.unauthorized = UnauthorizedHandler(self, username, \
                                                        request)
        return None

    def _basic_validate(self, request, auth, roles, parent):
        # Similar to BasicUserFolder.validate, except don't try
        # "If no authorization, only a user with a domain spec and no passwd 
        # or nobody can match" section, as this is real bad with large user
        # bases.

        # Only do basic authentication
        if not auth or string.lower(auth[:6])!='basic ':
            return None
        name,password=tuple(string.split(
            base64.decodestring(string.split(auth)[-1]), ':', 1))

        # Try to get user and authenticate.

        # Get from a miniature cache intended for use only
        # in basic user validation.
        now = time.time()
        gotFromCache = 0
        try:
            basicUserCache = self._v_basicUserCache
        except AttributeError:
            basicUserCache = {}
            self._v_basicUserCache = basicUserCache
        try:
            user, expiration = basicUserCache[name]
            gotFromCache = 1
            if expiration <= now:
                user = None
            else:
                user = user.__of__(self)
        except KeyError:
            user = None
        if user is None:
            gotFromCache = 0
            user = self.getUser(name)

        # Check authenticity of user and whether user is
        # allowed.
        if user is not None and user.allowed(parent, roles):
            if user.authenticate(password, request):
                if not gotFromCache:
                    LOG('GroupACL', INFO, \
                        '%s: User %s authenticated successfully (basic).' % \
                        (self.getTraversalPath(), user.getUserName()))
                    # Store in the cache for the same amount of time
                    # as cookieTimeout.
                    try:
                        timeout = int(self.getFlagValue('cookieTimeout'))
                    except TypeError:
                        timeout = 15  # Default to a few seconds.
                    basicUserCache[name] = (user.aq_base, now + timeout)
                return user
            else:
                try: del basicUserCache[name]
                except: pass
                LOG('GroupACL', INFO, \
                    '%s: User %s failed basic authentication.' % \
                    (self.getTraversalPath(), name))
        return None

    def validate(self, request, auth='', roles=None):
        # No docstring for a reason

        # Check from BasicUserFolder.validate. Who knows
        # what it does except that in some circumstances it will deny
        # access (better safe than sorry)
        if roles is _what_not_even_god_should_do:
            _debug('_what_not_even_god_should_do request')
            request.response.notFoundError()

	parents = request.get('PARENTS',[])
	if not parents:
	    parent = self.aq_parent
	else:
	    parent = parents[0]

        if self.getFlagValue('cookieMode'):
            u = self._cookie_validate(request, auth, roles, parent)
        else:
            u = self._basic_validate(request, auth, roles, parent)

        if u is None:
	    # If Anonymous has required permissions, no need to login.
	    # Don't check unless we are the top acl_users folder, as a higher
	    # acl_users folder might still validate us and give higher privs.
	    if self._isTop() and self._nobody.allowed(parent, roles):
		anon = self._nobody.__of__(self)
		return anon
            _debug("Returning None from %s validate (roles=%s)" \
                % (self.getTraversalPath(),str(roles)))
        else:
            _debug("Returning %s from %s validate (roles=%s)" \
                % (u.getUserName(),self.getTraversalPath(),str(roles)))
        return u

    def __findLoginPage(self, name):
        '''Locates the given page.  The page may come from a different
        source depending on ACLManager flags.
        '''
        if self.getFlagValue('allowLocalLoginPages') and \
           hasattr(self.aq_parent, name):
            return getattr(self.aq_parent, name)
        else:
            pageContext = locateTrueParent(self.getACLManager())
            if hasattr(pageContext.aq_base, name):
                return getattr(pageContext, name)
            else:
                return None

    def _renderLoginPage(self, request):
        '''Renders the login page.'''
        loginPage = self.__findLoginPage('docLogin')
        if loginPage is None:
            return 'Unable to log in: no docLogin page found.'
        loginSuccessPage = self.__findLoginPage('docLoginSuccess')
        title = self.title
        if len(title) < 1:
            title = self.aq_parent.title
        if loginSuccessPage is not None:
            return loginPage(client=self.aq_parent, REQUEST=request, \
                             folderTitle=title, loginSuccessURL= \
                             self.absolute_url() + '/loginSuccess', \
                             loginDestination = request['PATH_INFO'])
        else:
            return loginPage(client=self.aq_parent, REQUEST=request, \
                             folderTitle=title, loginSuccessURL= \
                             request['PATH_INFO'])

    def _killcookie(self, request):
        '''Helper method to log in and log out.'''
        if not self.getFlagValue('cookieMode'):
            raise 'GroupACLError', \
                  'Login and logout only work in cookie mode'
        if request.cookies.has_key(cname):
            # Pull username & hash from cookie
            _privateCookieJar.deleteCookie(request, cname)            
        request['RESPONSE'].expireCookie(cname,path='/')
 
    def login(self, REQUEST):
        '''Causes a login to occur immediately. Not normally used.'''
        self._killcookie(REQUEST)
        return self._renderLoginPage(REQUEST)
 
    def logout(self, REQUEST):
        '''Causes a logout to occur immediately. Might not be used.'''
        self._killcookie(REQUEST)
        logoutPage = self.__findLoginPage('docLogout')
        if logoutPage is None:
            raise 'GroupACLError', 'Object docLogout not found.'
        return logoutPage(self, REQUEST)

    def loginSuccess(self, REQUEST):
        '''May be called upon successful login. Not normally used.'''
        self._killcookie(REQUEST)
        loginSuccessPage = self.__findLoginPage('docLoginSuccess')
        if loginSuccessPage is None:
            raise 'GroupACLError', 'Object docLoginSuccess not found.'
        return loginSuccessPage(self, REQUEST)


# A special nobody object is needed in order to pass the
# watermark test.
_tempNobody = AccessControl.User.User('Anonymous User','',('Anonymous',), [])
try:
    _tempNobody._v__marker__ = AccessControl.User._marker
except:
    # Pre-Zope 2.1.5
    pass

class UnauthorizedHandler:
    '''A replacement for the unauthorized() method.
    '''
    cookie_aware = 1

    def __init__(self, groupACL, username, request):
        self.groupACL = groupACL
        self.username = username
        self.request = request

    def __call__(self, *args, **kw):
        '''Renders the login page when the user fails
        authentication checks.
        '''
        if self.username is not None:
            LOG('GroupACL', INFO, \
                '%s: User %s failed cookie authentication.' % \
                (self.groupACL.getTraversalPath(), self.username))
        request = self.request
        request.set('isIndirectLogin', 1)
        if not request.has_key('AUTHENTICATED_USER'):
            # Provide a temporary user account while showing
            # the login page.
            request.set('AUTHENTICATED_USER', \
                        _tempNobody.__of__(self.groupACL))
        pg = self.groupACL._renderLoginPage(request)
        raise 'LoginRequired', str(pg)
